@@ -2,7 +2,7 @@
 	node.c (09.10.09)
 	exFAT file system implementation library.
 
-	Copyright (C) 2009, 2010  Andrew Nayenko
+	Copyright (C) 2010-2013  Andrew Nayenko
 
 	This program is free software: you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -86,8 +86,8 @@ static int opendir(struct exfat* ef, const struct exfat_node* dir,
 		exfat_error("out of memory");
 		return -ENOMEM;
 	}
-	exfat_read_raw(it->chunk, CLUSTER_SIZE(*ef->sb),
-			exfat_c2o(ef, it->cluster), ef->fd);
+	exfat_pread(ef->dev, it->chunk, CLUSTER_SIZE(*ef->sb),
+			exfat_c2o(ef, it->cluster));
 	return 0;
 }
 
@@ -108,20 +108,19 @@ static int fetch_next_entry(struct exfat* ef, const struct exfat_node* parent,
 	/* fetch the next cluster if needed */
 	if ((it->offset & (CLUSTER_SIZE(*ef->sb) - 1)) == 0)
 	{
+		/* reached the end of directory; the caller should check this
+		   condition too */
 		if (it->offset >= parent->size)
-		{
-			exfat_error("missing EOD entry (0x%"PRIx64", 0x%"PRIx64")",
-					it->offset, parent->size);
-			return 1;
-		}
+			return 0;
 		it->cluster = exfat_next_cluster(ef, parent, it->cluster);
 		if (CLUSTER_INVALID(it->cluster))
 		{
-			exfat_error("invalid cluster while reading directory");
+			exfat_error("invalid cluster 0x%x while reading directory",
+					it->cluster);
 			return 1;
 		}
-		exfat_read_raw(it->chunk, CLUSTER_SIZE(*ef->sb),
-				exfat_c2o(ef, it->cluster), ef->fd);
+		exfat_pread(ef->dev, it->chunk, CLUSTER_SIZE(*ef->sb),
+				exfat_c2o(ef, it->cluster));
 	}
 	return 0;
 }
@@ -184,26 +183,25 @@ static int readdir(struct exfat* ef, const struct exfat_node* parent,
 	le16_t* namep = NULL;
 	uint16_t reference_checksum = 0;
 	uint16_t actual_checksum = 0;
+	uint64_t real_size = 0;
 
 	*node = NULL;
 
 	for (;;)
 	{
-		/* every directory (even empty one) occupies at least one cluster and
-		   must contain EOD entry */
-		entry = get_entry_ptr(ef, it);
-
-		switch (entry->type)
+		if (it->offset >= parent->size)
 		{
-		case EXFAT_ENTRY_EOD:
 			if (continuations != 0)
 			{
-				exfat_error("expected %hhu continuations before EOD",
-						continuations);
+				exfat_error("expected %hhu continuations", continuations);
 				goto error;
 			}
 			return -ENOENT; /* that's OK, means end of directory */
+		}
 
+		entry = get_entry_ptr(ef, it);
+		switch (entry->type)
+		{
 		case EXFAT_ENTRY_FILE:
 			if (continuations != 0)
 			{
@@ -250,18 +248,7 @@ static int readdir(struct exfat* ef, const struct exfat_node* parent,
 			}
 			init_node_meta2(*node, meta2);
 			actual_checksum = exfat_add_checksum(entry, actual_checksum);
-			/* There are two fields that contain file size. Maybe they plan
-			   to add compression support in the future and one of those
-			   fields is visible (uncompressed) size and the other is real
-			   (compressed) size. Anyway, currently it looks like exFAT does
-			   not support compression and both fields must be equal. */
-			if (le64_to_cpu(meta2->real_size) != (*node)->size)
-			{
-				exfat_error("real size does not equal to size "
-						"(%"PRIu64" != %"PRIu64")",
-						le64_to_cpu(meta2->real_size), (*node)->size);
-				goto error;
-			}
+			real_size = le64_to_cpu(meta2->real_size);
 			/* empty files must be marked as non-contiguous */
 			if ((*node)->size == 0 && (meta2->flags & EXFAT_FLAG_CONTIGUOUS))
 			{
@@ -273,11 +260,8 @@ static int readdir(struct exfat* ef, const struct exfat_node* parent,
 			if (((*node)->flags & EXFAT_ATTRIB_DIR) &&
 				(*node)->size % CLUSTER_SIZE(*ef->sb) != 0)
 			{
-				char buffer[EXFAT_NAME_MAX + 1];
-
-				exfat_get_name(*node, buffer, EXFAT_NAME_MAX);
-				exfat_error("directory `%s' has invalid size %"PRIu64" bytes",
-						buffer, (*node)->size);
+				exfat_error("directory has invalid size %"PRIu64" bytes",
+						(*node)->size);
 				goto error;
 			}
 			--continuations;
@@ -296,10 +280,34 @@ static int readdir(struct exfat* ef, const struct exfat_node* parent,
 			namep += EXFAT_ENAME_MAX;
 			if (--continuations == 0)
 			{
+				/*
+				   There are two fields that contain file size. Maybe they
+				   plan to add compression support in the future and one of
+				   those fields is visible (uncompressed) size and the other
+				   is real (compressed) size. Anyway, currently it looks like
+				   exFAT does not support compression and both fields must be
+				   equal.
+
+				   There is an exception though: pagefile.sys (its real_size
+				   is always 0).
+				*/
+				if (real_size != (*node)->size)
+				{
+					char buffer[EXFAT_NAME_MAX + 1];
+
+					exfat_get_name(*node, buffer, EXFAT_NAME_MAX);
+					exfat_error("`%s' real size does not equal to size "
+							"(%"PRIu64" != %"PRIu64")", buffer,
+							real_size, (*node)->size);
+					goto error;
+				}
 				if (actual_checksum != reference_checksum)
 				{
-					exfat_error("invalid checksum (0x%hx != 0x%hx)",
-							actual_checksum, reference_checksum);
+					char buffer[EXFAT_NAME_MAX + 1];
+
+					exfat_get_name(*node, buffer, EXFAT_NAME_MAX);
+					exfat_error("`%s' has invalid checksum (0x%hx != 0x%hx)",
+							buffer, actual_checksum, reference_checksum);
 					goto error;
 				}
 				if (fetch_next_entry(ef, parent, it) != 0)
@@ -314,7 +322,8 @@ static int readdir(struct exfat* ef, const struct exfat_node* parent,
 			upcase = (const struct exfat_entry_upcase*) entry;
 			if (CLUSTER_INVALID(le32_to_cpu(upcase->start_cluster)))
 			{
-				exfat_error("invalid cluster in upcase table");
+				exfat_error("invalid cluster 0x%x in upcase table",
+						le32_to_cpu(upcase->start_cluster));
 				goto error;
 			}
 			if (le64_to_cpu(upcase->size) == 0 ||
@@ -335,15 +344,17 @@ static int readdir(struct exfat* ef, const struct exfat_node* parent,
 			}
 			ef->upcase_chars = le64_to_cpu(upcase->size) / sizeof(le16_t);
 
-			exfat_read_raw(ef->upcase, le64_to_cpu(upcase->size),
-					exfat_c2o(ef, le32_to_cpu(upcase->start_cluster)), ef->fd);
+			exfat_pread(ef->dev, ef->upcase, le64_to_cpu(upcase->size),
+					exfat_c2o(ef, le32_to_cpu(upcase->start_cluster)));
 			break;
 
 		case EXFAT_ENTRY_BITMAP:
 			bitmap = (const struct exfat_entry_bitmap*) entry;
-			if (CLUSTER_INVALID(le32_to_cpu(bitmap->start_cluster)))
+			ef->cmap.start_cluster = le32_to_cpu(bitmap->start_cluster);
+			if (CLUSTER_INVALID(ef->cmap.start_cluster))
 			{
-				exfat_error("invalid cluster in clusters bitmap");
+				exfat_error("invalid cluster 0x%x in clusters bitmap",
+						ef->cmap.start_cluster);
 				goto error;
 			}
 			ef->cmap.size = le32_to_cpu(ef->sb->cluster_count) -
@@ -355,7 +366,6 @@ static int readdir(struct exfat* ef, const struct exfat_node* parent,
 						le64_to_cpu(bitmap->size), (ef->cmap.size + 7) / 8);
 				goto error;
 			}
-			ef->cmap.start_cluster = le32_to_cpu(bitmap->start_cluster);
 			/* FIXME bitmap can be rather big, up to 512 MB */
 			ef->cmap.chunk_size = ef->cmap.size;
 			ef->cmap.chunk = malloc(le64_to_cpu(bitmap->size));
@@ -367,8 +377,8 @@ static int readdir(struct exfat* ef, const struct exfat_node* parent,
 				goto error;
 			}
 
-			exfat_read_raw(ef->cmap.chunk, le64_to_cpu(bitmap->size),
-					exfat_c2o(ef, ef->cmap.start_cluster), ef->fd);
+			exfat_pread(ef->dev, ef->cmap.chunk, le64_to_cpu(bitmap->size),
+					exfat_c2o(ef, ef->cmap.start_cluster));
 			break;
 
 		case EXFAT_ENTRY_LABEL:
@@ -447,17 +457,40 @@ int exfat_cache_directory(struct exfat* ef, struct exfat_node* dir)
 	return 0;
 }
 
+static void tree_attach(struct exfat_node* dir, struct exfat_node* node)
+{
+	node->parent = dir;
+	if (dir->child)
+	{
+		dir->child->prev = node;
+		node->next = dir->child;
+	}
+	dir->child = node;
+}
+
+static void tree_detach(struct exfat_node* node)
+{
+	if (node->prev)
+		node->prev->next = node->next;
+	else /* this is the first node in the list */
+		node->parent->child = node->next;
+	if (node->next)
+		node->next->prev = node->prev;
+	node->parent = NULL;
+	node->prev = NULL;
+	node->next = NULL;
+}
+
 static void reset_cache(struct exfat* ef, struct exfat_node* node)
 {
-	struct exfat_node* child;
-	struct exfat_node* next;
-
-	for (child = node->child; child; child = next)
+	while (node->child)
 	{
-		reset_cache(ef, child);
-		next = child->next;
-		free(child);
+		struct exfat_node* p = node->child;
+		reset_cache(ef, p);
+		tree_detach(p);
+		free(p);
 	}
+	node->flags &= ~EXFAT_ATTRIB_CACHED;
 	if (node->references != 0)
 	{
 		char buffer[EXFAT_NAME_MAX + 1];
@@ -465,10 +498,8 @@ static void reset_cache(struct exfat* ef, struct exfat_node* node)
 		exfat_warn("non-zero reference counter (%d) for `%s'",
 				node->references, buffer);
 	}
-	while (node->references--)
+	while (node->references)
 		exfat_put_node(ef, node);
-	node->child = NULL;
-	node->flags &= ~EXFAT_ATTRIB_CACHED;
 }
 
 void exfat_reset_cache(struct exfat* ef)
@@ -505,14 +536,14 @@ void exfat_flush_node(struct exfat* ef, struct exfat_node* node)
 	next_entry(ef, node->parent, &cluster, &offset);
 	meta2_offset = co2o(ef, cluster, offset);
 
-	exfat_read_raw(&meta1, sizeof(meta1), meta1_offset, ef->fd);
+	exfat_pread(ef->dev, &meta1, sizeof(meta1), meta1_offset);
 	if (meta1.type != EXFAT_ENTRY_FILE)
 		exfat_bug("invalid type of meta1: 0x%hhx", meta1.type);
 	meta1.attrib = cpu_to_le16(node->flags);
 	exfat_unix2exfat(node->mtime, &meta1.mdate, &meta1.mtime, &meta1.mtime_cs);
 	exfat_unix2exfat(node->atime, &meta1.adate, &meta1.atime, NULL);
 
-	exfat_read_raw(&meta2, sizeof(meta2), meta2_offset, ef->fd);
+	exfat_pread(ef->dev, &meta2, sizeof(meta2), meta2_offset);
 	if (meta2.type != EXFAT_ENTRY_FILE_INFO)
 		exfat_bug("invalid type of meta2: 0x%hhx", meta2.type);
 	meta2.size = meta2.real_size = cpu_to_le64(node->size);
@@ -525,8 +556,8 @@ void exfat_flush_node(struct exfat* ef, struct exfat_node* node)
 
 	meta1.checksum = exfat_calc_checksum(&meta1, &meta2, node->name);
 
-	exfat_write_raw(&meta1, sizeof(meta1), meta1_offset, ef->fd);
-	exfat_write_raw(&meta2, sizeof(meta2), meta2_offset, ef->fd);
+	exfat_pwrite(ef->dev, &meta1, sizeof(meta1), meta1_offset);
+	exfat_pwrite(ef->dev, &meta2, sizeof(meta2), meta2_offset);
 
 	node->flags &= ~EXFAT_ATTRIB_DIRTY;
 }
@@ -539,53 +570,27 @@ static void erase_entry(struct exfat* ef, struct exfat_node* node)
 	uint8_t entry_type;
 
 	entry_type = EXFAT_ENTRY_FILE & ~EXFAT_ENTRY_VALID;
-	exfat_write_raw(&entry_type, 1, co2o(ef, cluster, offset), ef->fd);
+	exfat_pwrite(ef->dev, &entry_type, 1, co2o(ef, cluster, offset));
 
 	next_entry(ef, node->parent, &cluster, &offset);
 	entry_type = EXFAT_ENTRY_FILE_INFO & ~EXFAT_ENTRY_VALID;
-	exfat_write_raw(&entry_type, 1, co2o(ef, cluster, offset), ef->fd);
+	exfat_pwrite(ef->dev, &entry_type, 1, co2o(ef, cluster, offset));
 
 	while (name_entries--)
 	{
 		next_entry(ef, node->parent, &cluster, &offset);
 		entry_type = EXFAT_ENTRY_FILE_NAME & ~EXFAT_ENTRY_VALID;
-		exfat_write_raw(&entry_type, 1, co2o(ef, cluster, offset), ef->fd);
+		exfat_pwrite(ef->dev, &entry_type, 1, co2o(ef, cluster, offset));
 	}
 }
 
-static void tree_detach(struct exfat_node* node)
-{
-	if (node->prev)
-		node->prev->next = node->next;
-	else /* this is the first node in the list */
-		node->parent->child = node->next;
-	if (node->next)
-		node->next->prev = node->prev;
-	node->parent = NULL;
-	node->prev = NULL;
-	node->next = NULL;
-}
-
-static void tree_attach(struct exfat_node* dir, struct exfat_node* node)
-{
-	node->parent = dir;
-	if (dir->child)
-	{
-		dir->child->prev = node;
-		node->next = dir->child;
-	}
-	dir->child = node;
-}
-
 static int shrink_directory(struct exfat* ef, struct exfat_node* dir,
 		off_t deleted_offset)
 {
 	const struct exfat_node* node;
 	const struct exfat_node* last_node;
-	uint64_t entries = 1; /* a directory always has at leat 1 entry (EOD) */
+	uint64_t entries = 0;
 	uint64_t new_size;
-	struct exfat_entry eod;
-	off_t eod_offset;
 	int rc;
 
 	if (!(dir->flags & EXFAT_ATTRIB_DIR))
@@ -618,21 +623,13 @@ static int shrink_directory(struct exfat* ef, struct exfat_node* dir,
 
 	new_size = DIV_ROUND_UP(entries * sizeof(struct exfat_entry),
 				 CLUSTER_SIZE(*ef->sb)) * CLUSTER_SIZE(*ef->sb);
+	if (new_size == 0) /* directory always has at least 1 cluster */
+		new_size = CLUSTER_SIZE(*ef->sb);
 	if (new_size == dir->size)
 		return 0;
 	rc = exfat_truncate(ef, dir, new_size);
 	if (rc != 0)
 		return rc;
-
-	/* put EOD entry at the end of the last cluster */
-	memset(&eod, 0, sizeof(eod));
-	eod_offset = new_size - sizeof(struct exfat_entry);
-	if (last_node)
-		exfat_write_raw(&eod, sizeof(eod),
-				co2o(ef, last_node->entry_cluster, eod_offset), ef->fd);
-	else
-		exfat_write_raw(&eod, sizeof(eod),
-				co2o(ef, dir->start_cluster, eod_offset), ef->fd);
 	return 0;
 }
 
@@ -698,24 +695,22 @@ static int find_slot(struct exfat* ef, struct exfat_node* dir,
 			*offset = it.offset;
 		}
 		entry = get_entry_ptr(ef, &it);
-		if (entry->type == EXFAT_ENTRY_EOD)
-		{
-			rc = grow_directory(ef, dir,
-					it.offset + sizeof(struct exfat_entry), /* actual size */
-					(subentries - contiguous) * sizeof(struct exfat_entry));
-			if (rc != 0)
-			{
-				closedir(&it);
-				return rc;
-			}
-			break;
-		}
 		if (entry->type & EXFAT_ENTRY_VALID)
 			contiguous = 0;
 		else
 			contiguous++;
 		if (contiguous == subentries)
 			break;	/* suitable slot is found */
+		if (it.offset + sizeof(struct exfat_entry) >= dir->size)
+		{
+			rc = grow_directory(ef, dir, dir->size,
+					(subentries - contiguous) * sizeof(struct exfat_entry));
+			if (rc != 0)
+			{
+				closedir(&it);
+				return rc;
+			}
+		}
 		if (fetch_next_entry(ef, dir, &it) != 0)
 		{
 			closedir(&it);
@@ -751,8 +746,7 @@ static int write_entry(struct exfat* ef, struct exfat_node* dir,
 			&meta1.crtime_cs);
 	meta1.adate = meta1.mdate = meta1.crdate;
 	meta1.atime = meta1.mtime = meta1.crtime;
-	/* crtime_cs and mtime_cs contain addition to the time in centiseconds;
-	   just ignore those fields because we operate with 2 sec resolution */
+	meta1.mtime_cs = meta1.crtime_cs; /* there is no atime_cs */
 
 	memset(&meta2, 0, sizeof(meta2));
 	meta2.type = EXFAT_ENTRY_FILE_INFO;
@@ -763,17 +757,17 @@ static int write_entry(struct exfat* ef, struct exfat_node* dir,
 
 	meta1.checksum = exfat_calc_checksum(&meta1, &meta2, node->name);
 
-	exfat_write_raw(&meta1, sizeof(meta1), co2o(ef, cluster, offset), ef->fd);
+	exfat_pwrite(ef->dev, &meta1, sizeof(meta1), co2o(ef, cluster, offset));
 	next_entry(ef, dir, &cluster, &offset);
-	exfat_write_raw(&meta2, sizeof(meta2), co2o(ef, cluster, offset), ef->fd);
+	exfat_pwrite(ef->dev, &meta2, sizeof(meta2), co2o(ef, cluster, offset));
 	for (i = 0; i < name_entries; i++)
 	{
 		struct exfat_entry_name name_entry = {EXFAT_ENTRY_FILE_NAME, 0};
 		memcpy(name_entry.name, node->name + i * EXFAT_ENAME_MAX,
 				EXFAT_ENAME_MAX * sizeof(le16_t));
 		next_entry(ef, dir, &cluster, &offset);
-		exfat_write_raw(&name_entry, sizeof(name_entry),
-				co2o(ef, cluster, offset), ef->fd);
+		exfat_pwrite(ef->dev, &name_entry, sizeof(name_entry),
+				co2o(ef, cluster, offset));
 	}
 
 	init_node_meta1(node, &meta1);
@@ -855,11 +849,11 @@ static void rename_entry(struct exfat* ef, struct exfat_node* dir,
 	const int name_entries = DIV_ROUND_UP(name_length, EXFAT_ENAME_MAX);
 	int i;
 
-	exfat_read_raw(&meta1, sizeof(meta1), co2o(ef, old_cluster, old_offset),
-			ef->fd);
+	exfat_pread(ef->dev, &meta1, sizeof(meta1),
+			co2o(ef, old_cluster, old_offset));
 	next_entry(ef, node->parent, &old_cluster, &old_offset);
-	exfat_read_raw(&meta2, sizeof(meta2), co2o(ef, old_cluster, old_offset),
-			ef->fd);
+	exfat_pread(ef->dev, &meta2, sizeof(meta2),
+			co2o(ef, old_cluster, old_offset));
 	meta1.continuations = 1 + name_entries;
 	meta2.name_hash = exfat_calc_name_hash(ef, name);
 	meta2.name_length = name_length;
@@ -870,11 +864,11 @@ static void rename_entry(struct exfat* ef, struct exfat_node* dir,
 	node->entry_cluster = new_cluster;
 	node->entry_offset = new_offset;
 
-	exfat_write_raw(&meta1, sizeof(meta1), co2o(ef, new_cluster, new_offset),
-			ef->fd);
+	exfat_pwrite(ef->dev, &meta1, sizeof(meta1),
+			co2o(ef, new_cluster, new_offset));
 	next_entry(ef, dir, &new_cluster, &new_offset);
-	exfat_write_raw(&meta2, sizeof(meta2), co2o(ef, new_cluster, new_offset),
-			ef->fd);
+	exfat_pwrite(ef->dev, &meta2, sizeof(meta2),
+			co2o(ef, new_cluster, new_offset));
 
 	for (i = 0; i < name_entries; i++)
 	{
@@ -882,8 +876,8 @@ static void rename_entry(struct exfat* ef, struct exfat_node* dir,
 		memcpy(name_entry.name, name + i * EXFAT_ENAME_MAX,
 				EXFAT_ENAME_MAX * sizeof(le16_t));
 		next_entry(ef, dir, &new_cluster, &new_offset);
-		exfat_write_raw(&name_entry, sizeof(name_entry),
-				co2o(ef, new_cluster, new_offset), ef->fd);
+		exfat_pwrite(ef->dev, &name_entry, sizeof(name_entry),
+				co2o(ef, new_cluster, new_offset));
 	}
 
 	memcpy(node->name, name, (EXFAT_NAME_MAX + 1) * sizeof(le16_t));
@@ -911,29 +905,52 @@ int exfat_rename(struct exfat* ef, const char* old_path, const char* new_path)
 		exfat_put_node(ef, node);
 		return rc;
 	}
+
+	/* check that target is not a subdirectory of the source */
+	if (node->flags & EXFAT_ATTRIB_DIR)
+	{
+		struct exfat_node* p;
+
+		for (p = dir; p; p = p->parent)
+			if (node == p)
+			{
+				if (existing != NULL)
+					exfat_put_node(ef, existing);
+				exfat_put_node(ef, dir);
+				exfat_put_node(ef, node);
+				return -EINVAL;
+			}
+	}
+
 	if (existing != NULL)
 	{
-		if (existing->flags & EXFAT_ATTRIB_DIR)
+		/* remove target if it's not the same node as source */
+		if (existing != node)
 		{
-			if (node->flags & EXFAT_ATTRIB_DIR)
-				rc = exfat_rmdir(ef, existing);
+			if (existing->flags & EXFAT_ATTRIB_DIR)
+			{
+				if (node->flags & EXFAT_ATTRIB_DIR)
+					rc = exfat_rmdir(ef, existing);
+				else
+					rc = -ENOTDIR;
+			}
 			else
-				rc = -ENOTDIR;
+			{
+				if (!(node->flags & EXFAT_ATTRIB_DIR))
+					rc = exfat_unlink(ef, existing);
+				else
+					rc = -EISDIR;
+			}
+			exfat_put_node(ef, existing);
+			if (rc != 0)
+			{
+				exfat_put_node(ef, dir);
+				exfat_put_node(ef, node);
+				return rc;
+			}
 		}
 		else
-		{
-			if (!(node->flags & EXFAT_ATTRIB_DIR))
-				rc = exfat_unlink(ef, existing);
-			else
-				rc = -EISDIR;
-		}
-		exfat_put_node(ef, existing);
-		if (rc != 0)
-		{
-			exfat_put_node(ef, dir);
-			exfat_put_node(ef, node);
-			return rc;
-		}
+			exfat_put_node(ef, existing);
 	}
 
 	rc = find_slot(ef, dir, &cluster, &offset,
@@ -978,7 +995,6 @@ static int find_label(struct exfat* ef, cluster_t* cluster, off_t* offset)
 {
 	struct iterator it;
 	int rc;
-	const struct exfat_entry* entry;
 
 	rc = opendir(ef, ef->root, &it);
 	if (rc != 0)
@@ -986,14 +1002,13 @@ static int find_label(struct exfat* ef, cluster_t* cluster, off_t* offset)
 
 	for (;;)
 	{
-		entry = get_entry_ptr(ef, &it);
-
-		if (entry->type == EXFAT_ENTRY_EOD)
+		if (it.offset >= ef->root->size)
 		{
 			closedir(&it);
 			return -ENOENT;
 		}
-		if (entry->type == EXFAT_ENTRY_LABEL)
+
+		if (get_entry_ptr(ef, &it)->type == EXFAT_ENTRY_LABEL)
 		{
 			*cluster = it.cluster;
 			*offset = it.offset;
@@ -1034,7 +1049,7 @@ int exfat_set_label(struct exfat* ef, const char* label)
 	if (entry.length == 0)
 		entry.type ^= EXFAT_ENTRY_VALID;
 
-	exfat_write_raw(&entry, sizeof(struct exfat_entry_label),
-			co2o(ef, cluster, offset), ef->fd);
+	exfat_pwrite(ef->dev, &entry, sizeof(struct exfat_entry_label),
+			co2o(ef, cluster, offset));
 	return 0;
 }
